/* lspanal -- speech waveform to line spectral pairs + residual */

/* M.A.Huckvale - University College London */

/* version 1.0 - November 2006 */

#undef IAG

#define PROGNAME "lspanal"
#define PROGVERS "1.0"
char	*progname=PROGNAME;

/*--------------------------------------------------------------------------*/
/**MAN
.TH LSPANAL SFS1 UCL
.SH NAME
lspanal - code speech waveform with line spectral pairs + residual
.SH SYNOPSIS
.B lspanal
(-w window) (-s step) (-n ncoeffs) (-i item) (-t) (-g) file
.SH DESCRIPTION
.I lspanal
performs a windowed LPC analysis on sections of a speech waveform and
converts the predictor polynomial to line spectral pairs (frequencies).
The LSPs and a residual are stored.
.PP
.I Options
and their meanings are:
.TP 11
.B -I
Identify program and version number.
.TP 11
.BI -i item
Select input item number.
.TP 11
.BI -w window
Select analysis window size in milliseconds.  Default: 20ms.
.TP 11
.BI -s step
Select analysis window step size in milliseconds. Default: 10ms.
.TP 11
.BI -n ncoeff
Specify number of predictor coefficients to use.  Default is (2+2*sample_rate/1000).
.TP 11
.B -t
Perform pitch-synchronous analysis.  Requires TX item in input file.
.TP 11
.B -g
Calculate and save group delay function from residual as track.
.SH INPUT ITEMS
.IP 1.xx 11
Any speech waveform
.IP 3.xx 11
(Optional) Excitation markers.
.SH OUTPUT ITEMS
.IP 10 11
Line spectral frequencies (in fractions of sample rate).
.SH VERSION/AUTHOR
.nf
1.0 - Mark Huckvale
*/
/*--------------------------------------------------------------------------*/

/* standard definitions */

#include "SFSCONFG.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include "complex.h"
#include "filter.h"
#include "lsys.h"
#include "lsp.h"
#include "sfs.h"

#define MAXCOEFF	50	/* max # LPC coefficients */
#define	PREEMP		0.95	/* pre-emphasis coeff */
#define MAXWINDOW	22050	/* maximum analysis window size = 50ms @ 44.1kHz */

/* global data */
char		filename[SFSMAXFILENAME];	/* dbase file name */
struct item_header	spitem;		/* speech item header */
short			*sp;		/* speech buffer */
struct item_header	fxitem;
short		*fx;
struct item_header	txitem;		/* tx item header */
int			*tx;				/* tx data */
struct item_header	rspitem;		/* residual item header */
short			*rsp;		/* residual buffer */
struct item_header	pcitem;		/* LPC item header */
struct co_rec	*lpcrec;	/* buffer for LPC data */
int			ncoeff;				/* # LPC coefficients */
struct item_header	gtritem;		/* group delay item header */
float			*gtr;		/* group delay buffer */
double		witime=0.020;
double		sttime=0.010;
int			wisize;
int			stsize;
double		lsp[MAXCOEFF];
double		pc[MAXCOEFF];

/* tx windows */
struct txwin_rec {
	int		spos;
	int		snum;
	int		flag;			/* voicing */
} *txwin;
int txcnt;

/* median smoothing for fx */
void mediansmooth()
{
	short	window[5];
	short	winsrt[5];
	int		i,j,k;
	int		t;

	for (i=0;i<5;i++) window[i]=0;

	for (i=0;i<fxitem.numframes;i++) {
		for (j=1;j<5;j++) window[j-1]=window[j];
		window[4]=fx[i];

		/* sort samples */
		winsrt[0]=window[0];
		for (j=1;j<5;j++) {
			k=j;
			t=window[k];
			while ((k > 0)&&(t < winsrt[k-1])) {
				winsrt[k] = winsrt[k-1];
				k--;
			}
			winsrt[k]=t;
		}
		fx[i] = winsrt[2];
	}

}

/* join voiced regions with an interpolation */
void interpolate()
{
	double	sval,eval,gradient;
	int		start,stop;
	int		i,j;

	start=0;
	for (i=1;i<fxitem.numframes;i++) {
		if ((fx[i-1]==0)&&(fx[i]!=0)) {
			stop=i;
			sval = fx[start];
			if (sval==0) sval=fx[i];
			eval = fx[i];
			gradient = (eval-sval)/(stop-start);
			for (j=start;j<stop;j++)
				fx[j] = (short)(sval+gradient*(j-start));
		}
		else if ((fx[i-1]!=0)&&(fx[i]==0)) {
			start=i-1;
			sval=fx[start];
		}
	}
	i=fxitem.numframes-1;
	if (fx[i]==0) {
		for (j=start;j<=i;j++)
			fx[j] = (short)sval;
	}
}

/* make up an fx track from the tx track */
void makefx()
{
	double	t;
	int		i,idx,ofid,numf;
	int	maxtx=(int)(witime/txitem.frameduration);

	sfsheader(&fxitem,FX_TYPE,0,2,1,0.01,0.0,1,0,0);
	numf=fxitem.numframes = (int)(spitem.numframes*spitem.frameduration/fxitem.frameduration);
	fx = (short *)sfsbuffer(&fxitem,numf);
	memset(fx,0,numf*sizeof(short));

	t=0;
	for (i=0;i<txitem.numframes;i++) {
		if (tx[i]<maxtx) {
			idx=(int)(t/fxitem.frameduration);
			if ((0<=idx)&&(idx<numf))
				fx[idx]=(short)(1.0/(tx[i]*txitem.frameduration));
		}
		t += tx[i]*txitem.frameduration;
	}

	mediansmooth();
	interpolate();

	/* save to file */
#ifdef IAG
	sprintf(fxitem.history,"%s/FX(%d.%02d)",PROGNAME,txitem.datatype,txitem.subtype);
	ofid=sfschannel(filename,&fxitem);
	sfswrite(ofid,numf,fx);
#endif
}

int getwinsize(int pos)
{
	int idx=(int)(pos*txitem.frameduration/fxitem.frameduration);
	if (idx<0)
		return(int)((1.0/fx[0])/txitem.frameduration);
	else if (idx<fxitem.numframes)
		return(int)((1.0/fx[idx])/txitem.frameduration);
	else
		return(int)((1.0/fx[fxitem.numframes-1])/txitem.frameduration);
}

int findwin(int spos,int epos,int *win)
{
	int	i,cnt=0;
	int	w;
	int pos;
	double	factor;

	pos=spos;
	while (pos < epos) {
		w = getwinsize(pos);
		pos += w;
		cnt++;
	}
	if ((epos-pos) > w/2) {
		/* lengthen everything */
		pos -= w;
		cnt--;
		factor = (double)(epos-spos)/(pos-spos);
	}
	else {
		/* shorten everything */
		factor = (double)(epos-spos)/(pos-spos);
	}

	if (win) {
		pos=spos;
		for (i=0;i<cnt;i++) {
			win[i]=(int)(getwinsize(pos)*factor);
			pos += win[i];
		}
		win[cnt-1] += epos-pos;

#ifdef IAG
printf("win=");
for (i=0;i<cnt;i++) printf("%d,",win[i]);
printf("\n");
#endif

	}

	return(cnt);

}

/* calculate analysis window positions */
void calctxwin()
{
	int	i,j;
	int	num,txpos;
	int	mintx=(int)(2*ncoeff*spitem.frameduration/txitem.frameduration);
	int	maxtx=(int)(witime/txitem.frameduration);
	double	spend=spitem.numframes*spitem.frameduration;
	double	txlast,txthis,txnext;
	int		win[1000];

#ifdef IAG
	struct item_header txitem2;
	int	ofid;
#endif

	/* count how many windows we've got */
	txcnt=0;
	txpos=0;
	for (i=0;i<txitem.numframes;i++) {
		if (tx[i] <= maxtx)
			txcnt++;
		else {
			txcnt+=findwin(txpos,txpos+tx[i],NULL);
		}
		txpos += tx[i];
	}
	txcnt += findwin(txpos,(int)(spend/txitem.frameduration),NULL);

	/* allocate some memory */
	if ((txwin=(struct txwin_rec *)calloc(txcnt+100,sizeof(struct txwin_rec)))==NULL)
		error("could not get memory for buffer");

	/* calculate the window positions */
	txcnt=0;
	txlast=0;
	for (i=0;(i<txitem.numframes)&&(txlast<spend-witime);i++) {
		txthis = txlast + tx[i]*txitem.frameduration;
		if (i < txitem.numframes-1)
			txnext = txthis+tx[i+1]*txitem.frameduration;
		else
			txnext = spend;
		if (txnext > txthis+witime/2) txnext=txthis+witime/2;
		if (txthis > spend) txthis=spend;
		if (txnext > spend) txnext=spend;
		if (txlast > txthis) break;

		if (tx[i] <= maxtx) {
			txwin[txcnt].spos = (int)(txlast/spitem.frameduration);
			txwin[txcnt].snum = (int)((txnext-txlast)/spitem.frameduration);
			txwin[txcnt].flag = 1;
#ifdef IAG
printf("1 txlast=%g txthis=%g txnext=%g spos=%d snum=%d stot=%d\n",
	txlast,txthis,txnext,txwin[txcnt].spos,txwin[txcnt].snum,spitem.numframes);
#endif
			if ((txwin[txcnt].spos+txwin[txcnt].snum)>spitem.numframes)
				txwin[txcnt].snum = spitem.numframes-txwin[txcnt].spos;
			if (txwin[txcnt].snum > ncoeff) txcnt++;
		}
		else {
			num=findwin((int)(txlast/txitem.frameduration),(int)(txthis/txitem.frameduration),win);
			for (j=0;j<num-1;j++) {
				txwin[txcnt].spos = (int)(txlast/spitem.frameduration);
				txwin[txcnt].snum = (int)((win[j]+win[j+1])*txitem.frameduration/spitem.frameduration);
				txwin[txcnt].flag = 0;
#ifdef IAG
printf("2 txlast=%g txthis=%g txnext=%g spos=%d snum=%d stot=%d\n",
	txlast,txthis,txnext,txwin[txcnt].spos,txwin[txcnt].snum,spitem.numframes);
#endif
				if ((txwin[txcnt].spos+txwin[txcnt].snum)>spitem.numframes)
					txwin[txcnt].snum = spitem.numframes-txwin[txcnt].spos;
				if (txwin[txcnt].snum > ncoeff) txcnt++;
				txlast += win[j]*txitem.frameduration;
			}
			txwin[txcnt].spos = (int)(txlast/spitem.frameduration);
			txwin[txcnt].snum = (int)((txnext-txlast)/spitem.frameduration);
			txwin[txcnt].flag = 0;
#ifdef IAG
printf("3 txlast=%g txthis=%g txnext=%g spos=%d snum=%d stot=%d\n",
	txlast,txthis,txnext,txwin[txcnt].spos,txwin[txcnt].snum,spitem.numframes);
#endif
			if ((txwin[txcnt].spos+txwin[txcnt].snum)>spitem.numframes)
				txwin[txcnt].snum = spitem.numframes-txwin[txcnt].spos;
			if (txwin[txcnt].snum > ncoeff) txcnt++;
		}
		txlast=txthis;
	}

	if (txlast < (spend-witime)) {
		txthis=txnext=spend;
		num=findwin((int)(txlast/txitem.frameduration),(int)(txthis/txitem.frameduration),win);
		for (j=0;j<num-1;j++) {
			txwin[txcnt].spos = (int)(txlast/spitem.frameduration);
			txwin[txcnt].snum = (int)((win[j]+win[j+1])*txitem.frameduration/spitem.frameduration);
			txwin[txcnt].flag = 0;
#ifdef IAG
printf("4 txlast=%g txthis=%g txnext=%g spos=%d snum=%d stot=%d\n",
	txlast,txthis,txnext,txwin[txcnt].spos,txwin[txcnt].snum,spitem.numframes);
#endif
			if ((txwin[txcnt].spos+txwin[txcnt].snum)>spitem.numframes)
				txwin[txcnt].snum = spitem.numframes-txwin[txcnt].spos;
			if (txwin[txcnt].snum > ncoeff) txcnt++;
			txlast += win[j]*txitem.frameduration;
		}
		txwin[txcnt].spos = (int)(txlast/spitem.frameduration);
		txwin[txcnt].snum = (int)((txnext-txlast)/spitem.frameduration);
		txwin[txcnt].flag = 0;
#ifdef IAG
printf("5 txlast=%g txthis=%g txnext=%g spos=%d snum=%d stot=%d\n",
	txlast,txthis,txnext,txwin[txcnt].spos,txwin[txcnt].snum,spitem.numframes);
#endif
		if ((txwin[txcnt].spos+txwin[txcnt].snum)>spitem.numframes)
			txwin[txcnt].snum = spitem.numframes-txwin[txcnt].spos;
		if (txwin[txcnt].snum > ncoeff) txcnt++;
	}

	/* save to file */
#ifdef IAG
	sfsheader(&txitem2,TX_TYPE,0,4,1,spitem.frameduration,0.0,1,0,0);
	sprintf(txitem2.history,"%s/TX(%d.%02d)",PROGNAME,txitem.datatype,txitem.subtype);
	ofid=sfschannel(filename,&txitem2);
	for (i=1;i<txcnt;i++) {
		num=txwin[i].spos - txwin[i-1].spos;
		sfswrite(ofid,1,&num);
	}
#endif

}


/* calculate group delay from residual - as in Brookes et al IEEE Trans May 2006 */
void groupdelay(float *op,short *ip,int numf,int wlen)
{
	double	*win=(double *)calloc(wlen,sizeof(double));
	int		i,j;
	double	samp,arg;
	COMPLEX	x2;
	COMPLEX	sum;
	FILTER	*lpfilt;

	lpfilt=filter_design(FILTER_LOW_PASS,4,4000.0,1.0/spitem.frameduration,1.0/spitem.frameduration);
	filter_signal(lpfilt,ip,ip,numf);

	for (i=0;i<wlen;i++) win[i]=0.54-0.46*cos(2*M_PI*i/(wlen-1));

	for (i=wlen/2;i<numf-wlen/2-1;i++) {
		sum=CMake(0.0,0.0);
		for (j=0;j<wlen;j++) {
			samp=win[j]*ip[i-wlen/2+j];
			x2=CMake(samp*samp,0.0);
			sum=CAdd(sum,CMult(x2,CExp(CMake(0,M_PI*(2*j+1)/wlen))));
		}
		arg=CArg(sum);
		if (arg < 0) arg += 2*M_PI;
		op[i]=(float)(wlen*arg/(2*M_PI)-0.5-wlen/2);
	}
}

/* create random noise at quantisation level */
double quantnoise()
{
	int v=rand()%10000;
	return ((v-5000)/10000.0);
}

/* main program */
void main(argc,argv)
int	argc;
char	*argv[];
{
	/* option decoding */
	extern int	optind;		/* option index */
	extern char	*optarg;	/* option argument ptr */
	int		errflg = 0;	/* option error flag */
	int		c;		/* option switch */
	int		it;		/* item selections */
	char		*ty;
	char		*sptype="0";	/* default sub-type = last */
	char		*txtype="0";	/* default sub-type = last */
	int			dotx=0;
	/* file variables */
	int		fid,ofid,rfid,gfid;
	/* data variables */
	int		i,j,spos,scnt;
	double	omega,sum;
	double	*fsp,*frsp,*fwsp;
	int		dogroupdelay=0;

	/* decode switches */
	while ( (c = getopt(argc,argv,"Ii:w:s:n:gt")) != EOF ) switch (c) {
		case 'I' :	/* Identify */
			fprintf(stderr,"%s: Line spectral pair analysis V%s\n",PROGNAME,PROGVERS);
			exit(0);
			break;
		case 'i' :	/* specific item */
			if (itspec(optarg,&it,&ty) == 0) {
				if (it == SP_TYPE)
					sptype = ty;
				else if (it == TX_TYPE) {
					txtype = ty;
					dotx=1;
				}
				else
					error("unsuitable item specifier %s",optarg);
			}
			else
				error("illegal item specifier %s",optarg);
			break;
		case 'w' :	/* window time */
			witime = atof(optarg)/1E3;
			break;
		case 's' :	/* window step time */
			sttime = atof(optarg)/1E3;
			break;
		case 'n' :	/* specify # coeffs */
			ncoeff = atoi(optarg);
			if (ncoeff < 2)
				error("too few poles: '%s'",optarg);
			if (ncoeff > MAXCOEFF) ncoeff=MAXCOEFF;
			break;
		case 't':
			dotx=1;
			break;
		case 'g':
			dogroupdelay=1;
			break;
		case '?' :	/* unknown */
			errflg++;
	}
	if (errflg || (argc<2))
		error("usage: %s (-I) (-i item) (-w winsize[ms]) (-s stepsize[ms]) (-n numcoeff) (-t) file",PROGNAME);

	/* get filename */
	if (optind < argc)
		strcpy(filename,sfsfile(argv[optind]));
	else
		error("no database file specified",NULL);

	/* check file ok for writing */
	if ((fid=sfsopen(filename,"w",NULL)) < 0)
		error("access error on %s",filename);

	/* load Tx data */
	if (dotx) {
		/* locate TX data */
		if (!sfsitem(fid,TX_TYPE,txtype,&txitem))
			error("unable to find input TX item in %s",filename);
		tx=sfsbuffer(&txitem,txitem.numframes);
		sfsread(fid,0,txitem.numframes,tx);
	}

	/* locate speech data */
	if (!sfsitem(fid,SP_TYPE,sptype,&spitem))
		error("unable to find input speech item in %s",filename);

	if (dotx) makefx();

	/* create buffer for LPC coefficients */
	if (ncoeff==0) ncoeff = (int)(2 + 0.001/spitem.frameduration);
	if (ncoeff > MAXCOEFF) ncoeff=MAXCOEFF;
	wisize=(int)(0.5+witime/spitem.frameduration);
	stsize=(int)(0.5+sttime/spitem.frameduration);

	if (dotx) {
		sfsheader(&pcitem,PC_TYPE,-1,4,sfsstruct[PC_TYPE]/4 + ncoeff,
				spitem.frameduration,spitem.offset,0,0,1);
		sprintf(pcitem.history,"%s/PC(%d.%02d,%d.%02d;win=%g;step=%g,ncoeff=%d)",
			PROGNAME,
			spitem.datatype,spitem.subtype,
			txitem.datatype,txitem.subtype,
			witime,sttime,ncoeff);
	}
	else {
		sfsheader(&pcitem,PC_TYPE,-1,4,sfsstruct[PC_TYPE]/4 + ncoeff,
				spitem.frameduration,spitem.offset,wisize,wisize-stsize,0);
		sprintf(pcitem.history,"%s/PC(%d.%02d;win=%g,step=%g,ncoeff=%d)",
			PROGNAME,
			spitem.datatype,spitem.subtype,witime,sttime,ncoeff);
	}
	sprintf(pcitem.params,"format=LSP");
	if ((lpcrec=(struct co_rec *)sfsbuffer(&pcitem,1))==NULL)
		error("cannot get LPC buffer",NULL);

	/* create buffer for residual */
	sfsheader(&rspitem,LX_TYPE,0,2,1,spitem.frameduration,spitem.offset,1,0,0);
	if (dotx) {
		sprintf(rspitem.history,"%s/LX(%d.%02d,%d.%02d;win=%g,step=%g,ncoeff=%d)",
			PROGNAME,
			spitem.datatype,spitem.subtype,
			txitem.datatype,txitem.subtype,
			witime,sttime,ncoeff);
	}
	else {
		sprintf(rspitem.history,"%s/LX(%d.%02d;win=%g,step=%g,ncoeff=%d)",
			PROGNAME,
			spitem.datatype,spitem.subtype,witime,sttime,ncoeff);
	}

	/* open output channels */
	if ((ofid=sfschannel(filename,&pcitem))<0)
		error("could not open temporary file",NULL);

	/* get input signal */
	sp = (short *)sfsbuffer(&spitem,spitem.numframes);
	sfsread(fid,0,spitem.numframes,sp);
	/* pre-emp */
	for (j=spitem.numframes-1;j>0;j--) sp[j]=(short)(sp[j]-PREEMP*sp[j-1]);
	sp[0]=0;

	/* get data buffers */
	rsp = (short *)sfsbuffer(&rspitem,spitem.numframes);
	memset(rsp,0,sizeof(short)*spitem.numframes);
	fsp = (double *)calloc(MAXWINDOW,sizeof(double));
	frsp = (double *)calloc(MAXWINDOW,sizeof(double));
	fwsp = (double *)calloc(MAXWINDOW,sizeof(double));

	if (dotx) {
		/* pitch synchronous analysis */
		calctxwin();

		for (i=0;i<txcnt;i++) {
#ifdef IAG
printf("\n%d pos=%d num=%d\n",i,txwin[i].spos,txwin[i].snum);
#endif
			/* get speech */
			for (j=0;j<txwin[i].snum;j++) fsp[j]=(double)sp[txwin[i].spos+j];
			if (txwin[i].snum < wisize) {
				/* signal segment too small for good filter estimation */
				spos = txwin[i].spos + txwin[i].snum/2 - wisize/2;
				if (spos < 0) spos=0;
				scnt = wisize;
				if ((spos+scnt)>=spitem.numframes) scnt=spitem.numframes-spos;
				for (j=0;j<scnt;j++) fwsp[j]=(double)sp[spos+j] + quantnoise();
			}
			else {
				spos=txwin[i].spos;
				scnt=txwin[i].snum;
				for (j=0;j<txwin[i].snum;j++) fwsp[j]=(double)sp[txwin[i].spos+j] + quantnoise();
			}

			/* remove mean & window */
			sum=0;
			for (j=0;j<txwin[i].snum;j++) sum += fsp[j];
			sum /= txwin[i].snum;
			omega=(2*M_PI)/(txwin[i].snum-1);
			for (j=0;j<txwin[i].snum;j++)
				fsp[j] = (fsp[j]-sum)*(0.50-0.50*cos(omega*j));

			/* remove mean & window */
			sum=0;
			for (j=0;j<scnt;j++) sum += fsp[j];
			sum /= scnt;
			omega=(2*M_PI)/(scnt-1);
			for (j=0;j<scnt;j++)
				fwsp[j] = (fwsp[j]-sum)*(0.50-0.50*cos(omega*j));

			/* calculate RMS energy per sample */
			sum=0;
			for (j=0;j<txwin[i].snum;j++) sum += fsp[j]*fsp[j];
			sum = sqrt(sum/txwin[i].snum);

			/* perform LSP analysis on this window */
			lsp_auto_lpc(fwsp,scnt,pc,ncoeff);
			lsp_from_lpc(pc,lsp,ncoeff);
			lsp_residual_lpc(fsp,txwin[i].snum,pc,ncoeff,frsp);

			/* save */
			lpcrec->posn=txwin[i].spos;
			lpcrec->size=txwin[i].snum;
			lpcrec->flag=txwin[i].flag;
			lpcrec->mix=0;
			lpcrec->gain=(float)sum;
			for (j=0;j<ncoeff;j++)
				lpcrec->data[j]=(float)lsp[j];

			for (j=0;j<txwin[i].snum;j++)
				rsp[txwin[i].spos+j] = (short)(rsp[txwin[i].spos+j]+frsp[j]);

			/* write record */
			if (sfswrite(ofid,1,lpcrec) != 1)
				error("write error on output file",NULL);

			/* print progress */
			if (ttytest()) {
				printf("\rFrame %d/%d",i+1,txcnt);
				fflush(stdout);
			}
		}
	}
	else {
		/* process speech frame by frame */
		for (i=0;i<spitem.numframes-wisize;i+=stsize) {

			/* get speech */
			for (j=0;j<wisize;j++) fsp[j]=(double)sp[i+j];

			/* remove mean & window */
			sum=0;
			for (j=0;j<wisize;j++) sum += fsp[j];
			sum /= wisize;
			omega=(2*M_PI)/(wisize-1);
			for (j=0;j<wisize;j++)
				fsp[j] = (fsp[j]-sum)*(0.50-0.50*cos(omega*j));

			/* calculate RMS energy per sample */
			sum=0;
			for (j=0;j<wisize;j++) sum += fsp[j]*fsp[j];
			sum = sqrt(sum/wisize)*2*stsize/wisize;

			/* perform LSP analysis on this window */
			lsp_analysis(fsp,wisize,lsp,ncoeff,frsp);

			/* save */
			lpcrec->posn=i;
			lpcrec->size=wisize;
			lpcrec->flag=0;
			lpcrec->mix=0;
			lpcrec->gain=(float)sum;
			for (j=0;j<ncoeff;j++)
				lpcrec->data[j]=(float)lsp[j];

			for (j=0;j<wisize;j++)
				rsp[i+j] = (short)(rsp[i+j]+frsp[j]);

			/* write record */
			if (sfswrite(ofid,1,lpcrec) != 1)
				error("write error on output file",NULL);

			/* print progress */
			if (ttytest()) {
				printf("\rFrame %d/%d",i/wisize+1,spitem.numframes/wisize);
				fflush(stdout);
			}
		}
	}
	if (ttytest()) printf("\r                     \r");

	/* write residual */
	rfid=sfschannel(filename,&rspitem);
	sfswrite(rfid,spitem.numframes,rsp);

	if (dogroupdelay) {
		sfsheader(&gtritem,TR_TYPE,1,4,1,spitem.frameduration,spitem.offset,1,0,0);
		sprintf(gtritem.history,"%s/TR(%d.%02d;win=%g,step=%g,ncoeff=%d)",
			PROGNAME,
			spitem.datatype,spitem.subtype,witime,sttime,ncoeff);
		gtr=(float *)sfsbuffer(&gtritem,spitem.numframes);
		gfid=sfschannel(filename,&gtritem);
		groupdelay(gtr,rsp,spitem.numframes,(int)(0.008/spitem.frameduration));
		sfswrite(gfid,spitem.numframes,gtr);
	}

	/* update */
	if (!sfsupdate(filename) != 0)
		error("update error on %s",filename);

	/* that's all folks ... */
	exit(0);
}

